/*
 * main.cxx
 *
 * PWLib application source file for opengk
 *
 * Main program entry point.
 *
 * Copyright 2001 Equivalence
 *
 * $Log: main.cxx,v $
 * Revision 1.14  2002/05/31 08:33:16  robertj
 * Changed test oids to Equivalence private numbers
 *
 * Revision 1.13  2002/05/29 07:10:56  robertj
 * Added test of tokens in ACF.
 * Added http links to get at trace log file (if available).
 *
 * Revision 1.12  2002/05/07 06:51:23  robertj
 * Added application name to registered endpoint info.
 *
 * Revision 1.11  2002/05/07 06:16:58  robertj
 * Changed ARQ processing so rejects multiple ARQ with same call id, not this
 *   does NOT include the transport level retries (based on sequence number).
 *
 * Revision 1.10  2002/05/06 00:49:06  robertj
 * Sizeable rewrite of gatekeeper server code to make more bulletproof against
 *   multithreaded operation. Especially when using slow response/RIP feature.
 * Also changed the call indexing to use call id and direction as key.
 *
 * Revision 1.9  2002/03/01 05:37:35  robertj
 * Changed status call display to show both endpoints in call.
 *
 * Revision 1.8  2002/02/25 10:56:05  robertj
 * Added dynamically allocated port for H.323 channel listener.
 *
 * Revision 1.7  2002/02/17 23:51:18  robertj
 * Changed default listener port for future gk routing to 1721 instead of default.
 *
 * Revision 1.6  2002/02/04 05:20:03  robertj
 * Added call status display on HTTP.
 * Added test for slow ARQ
 *
 * Revision 1.5  2002/01/17 07:03:29  robertj
 * Added support for RFC2833 embedded DTMF in the RTP stream
 *
 * Revision 1.4  2001/12/22 04:33:54  robertj
 * Fixed ability to change listener port, thanks george.2.hare@bt.com
 *
 * Revision 1.3  2001/09/18 09:55:08  robertj
 * Changed default to have no username and password.
 *
 * Revision 1.2  2001/08/10 11:03:49  robertj
 * Major changes to H.235 support in RAS to support server.
 *
 * Revision 1.1  2001/08/03 04:48:57  robertj
 * First cut of a gatekeeper
 *
 */

#include <ptlib.h>
#include "main.h"
#include "custom.h"

#include <gsmcodec.h>

#include <ptclib/random.h>

//#define TEST_SLOW_ARQ
//#define TEST_TOKEN

PCREATE_PROCESS(OpenGk);

const WORD DefaultHTTPPort = 1719;

static const char LogLevelKey[] = "Log Level";
static const char AdminUserNameKey[] = "Username";
static const char AdminPasswordKey[] = "Password";
static const char HttpPortKey[] = "HTTP Port";
static const char LocalUserNameKey[] = "Local User Name";
static const char TypeOfServiceKey[] = "Type Of Service";
static const char GatekeeperIdentifierKey[] = "Gatekeeper Identifier";
static const char AvailableBandwidthKey[] = "Total Bandwidth";
static const char DefaultBandwidthKey[] = "Default Bandwidth Allocation";
static const char MaximumBandwidthKey[] = "Maximum Bandwidth Allocation";
static const char DefaultTimeToLiveKey[] = "Default Time To Live";
static const char OverwriteOnSameSignalAddressKey[] = "Overwrite EP On Same Signal Address";
static const char CanOnlyCallRegisteredEPKey[] = "Can Only Call Registered EP";
static const char CanOnlyAnswerRegisteredEPKey[] = "Can Only Answer Registered EP";
static const char AnswerCallPreGrantedARQKey[] = "Answer Call Pregranted ARQ";
static const char MakeCallPreGrantedARQKey[] = "Make Call Pregranted ARQ";
static const char IsGatekeeperRoutedKey[] = "Gatekeeper Routed";
static const char AliasCanBeHostNameKey[] = "Alias Can Be Host Name";
static const char UseH235Key[] = "Use H.235";
static const char UsernameKey[] = "Username";
static const char PasswordKey[] = "Password";

#define GATEWAY_LISTENERS_SECTION     "Gateway Listeners"
#define GATEKEEPER_LISTENERS_SECTION  "Gatekeeper Listeners"
#define LISTENER_INTERFACE_KEY        "Interface"

#define USERS_SECTION "Authentication"
#define USERS_KEY     "Credentials"


#define new PNEW


///////////////////////////////////////////////////////////////

OpenGk::OpenGk()
  : PHTTPServiceProcess(ProductInfo)
{
  endpoint = NULL;
}


BOOL OpenGk::OnStart()
{
  // Create the endpoint, or reinitialised if already exists
  endpoint = new MyEndPoint();


  // change to the default directory to the one containing the executable
  PDirectory exeDir = GetFile().GetDirectory();

#if defined(_WIN32) && defined(_DEBUG)
  // Special check to aid in using DevStudio for debugging.
  if (exeDir.Find("\\Debug\\") != P_MAX_INDEX)
    exeDir = exeDir.GetParent();
#endif
  exeDir.Change();

  httpNameSpace.AddResource(new PHTTPDirectory("data", "data"));
  httpNameSpace.AddResource(new PServiceHTTPDirectory("html", "html"));

  return PHTTPServiceProcess::OnStart();
}


void OpenGk::OnStop()
{
  delete endpoint;

  PHTTPServiceProcess::OnStop();
}



void OpenGk::OnControl()
{
  // This function get called when the Control menu item is selected in the
  // tray icon mode of the service.
  PStringStream url;
  url << "http://";

  PString host = PIPSocket::GetHostName();
  PIPSocket::Address addr;
  if (PIPSocket::GetHostAddress(host, addr))
    url << host;
  else
    url << "localhost";

  url << ':' << DefaultHTTPPort;

  PURL::OpenBrowser(url);
}


void OpenGk::OnConfigChanged()
{
}


PString OpenGk::GetPageGraphic()
{
  PHTML html(PHTML::InBody);
  html << "<CENTER>\r\n"
       << PHTML::Heading(1) << "OpenH323 Gatekeeper" << PHTML::Heading(1)
       << GetOSClass() << ' ' << GetOSName()
       << " Version " << GetVersion(TRUE) << PHTML::BreakLine()
       << ' ' << GetCompilationDate().AsString("d MMMM yyyy")
       << PHTML::BreakLine()
       << "by"
       << PHTML::Heading(3)
       << PHTML::HotLink(GetHomePage()) << GetManufacturer() << PHTML::HotLink()
       << PHTML::Heading(3)
       << PHTML::HotLink(PString("mailto:")+GetEMailAddress()) << GetEMailAddress() << PHTML::HotLink()
       << PHTML::Paragraph()
       << PHTML::HRule()
       << PHTML::Paragraph();

  return html;
}


BOOL OpenGk::Initialise(const char * initMsg)
{
  PConfig cfg("Parameters");

  // Sert log level as early as possible
  SetLogLevel((PSystemLog::Level)cfg.GetInteger(LogLevelKey, GetLogLevel()));
#if PTRACING
  if (GetLogLevel() >= PSystemLog::Warning)
    PTrace::SetLevel(GetLogLevel()-PSystemLog::Warning);
  else
    PTrace::SetLevel(0);
  PTrace::ClearOptions(PTrace::Timestamp);
  PTrace::SetOptions(PTrace::DateAndTime);
#endif

  // Get the HTTP basic authentication info
  PString adminUserName = cfg.GetString(AdminUserNameKey);
  PString adminPassword = PHTTPPasswordField::Decrypt(cfg.GetString(AdminPasswordKey));

  PHTTPSimpleAuth authority(GetName(), adminUserName, adminPassword);


  // Create the parameters URL page, and start adding fields to it
  PConfigPage * rsrc = new PConfigPage(*this, "Parameters", "Parameters", authority);

  // HTTP authentication username/password
  rsrc->Add(new PHTTPStringField(AdminUserNameKey, 25, adminUserName));
  rsrc->Add(new PHTTPPasswordField(AdminPasswordKey, 25, adminPassword));

  // Log level for messages
  rsrc->Add(new PHTTPIntegerField(LogLevelKey,
                                  PSystemLog::Fatal, PSystemLog::NumLogLevels-1,
                                  GetLogLevel(),
                                  "1=Fatal only, 2=Errors, 3=Warnings, 4=Info, 5=Debug"));

  // HTTP Port number to use.
  WORD httpPort = (WORD)cfg.GetInteger(HttpPortKey, DefaultHTTPPort);
  rsrc->Add(new PHTTPIntegerField(HttpPortKey, 1, 32767, httpPort));

  // Initialise the rest of the endpoint parameters
  endpoint->Initialise(cfg, rsrc);

  // Finished the resource to add, generate HTML for it and add to name space
  PServiceHTML html("System Parameters");
  rsrc->BuildHTML(html);
  httpNameSpace.AddResource(rsrc, PHTTPSpace::Overwrite);


  // Create the status page
  httpNameSpace.AddResource(new MainStatusPage(*this, authority), PHTTPSpace::Overwrite);


  // Add log file links
  if (!systemLogFileName && systemLogFileName != "-") {
    httpNameSpace.AddResource(new PHTTPFile("logfile.txt", systemLogFileName, authority));
    httpNameSpace.AddResource(new PHTTPTailFile("tail_logfile", systemLogFileName, authority));
  }


  //  create the home page
  static const char welcomeHtml[] = "welcome.html";
  if (PFile::Exists(welcomeHtml))
    httpNameSpace.AddResource(new PServiceHTTPFile(welcomeHtml, TRUE), PHTTPSpace::Overwrite);
  else {
    PHTML html;
    html << PHTML::Title("Welcome to OpenH323 Gatekeeper")
         << PHTML::Body()
         << GetPageGraphic()

         << PHTML::HotLink("Parameters") << "Parameters" << PHTML::HotLink()
         << PHTML::Paragraph()
         << PHTML::HotLink("Status") << "Status" << PHTML::HotLink()
         << PHTML::Paragraph();

    if (!systemLogFileName && systemLogFileName != "-")
      html << PHTML::HotLink("logfile.txt") << "Full Log File" << PHTML::HotLink()
           << PHTML::BreakLine()
           << PHTML::HotLink("tail_logfile") << "Tail Log File" << PHTML::HotLink()
           << PHTML::Paragraph();
 
    html << PHTML::HRule()
         << GetCopyrightText()
         << PHTML::Body();
    httpNameSpace.AddResource(new PServiceHTTPString("welcome.html", html), PHTTPSpace::Overwrite);
  }

  // set up the HTTP port for listening & start the first HTTP thread
  if (ListenForHTTP(httpPort))
    PSYSTEMLOG(Info, "Opened master socket for HTTP: " << httpListeningSocket->GetPort());
  else {
    PSYSTEMLOG(Fatal, "Cannot run without HTTP port: " << httpListeningSocket->GetErrorText());
    return FALSE;
  }

  PSYSTEMLOG(Info, "Service " << GetName() << ' ' << initMsg);
  return TRUE;
}


void OpenGk::Main()
{
  Suspend();
}


///////////////////////////////////////////////////////////////

MainStatusPage::MainStatusPage(OpenGk & _app, PHTTPAuthority & auth)
  : PServiceHTTPString("Status", auth), app(_app)
{
  PHTML html;

  html << PHTML::Title("OpenH323 Gatekeeper Status")
       << "<meta http-equiv=\"Refresh\" content=\"30\">\n"
       << PHTML::Body()
       << app.GetPageGraphic()

       << PHTML::Form("POST")

       << PHTML::TableStart("border=1")
       << PHTML::TableRow()
       << PHTML::TableHeader()
       << "&nbsp;End&nbsp;Point&nbsp;Identifier&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Call&nbsp;Signal&nbsp;Addresses&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Aliases&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Application&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Active&nbsp;Calls&nbsp;"
       << "<!--#macrostart EndPointStatus-->"
         << PHTML::TableRow()
         << PHTML::TableData()
         << "<!--#status EndPointIdentifier-->"
         << PHTML::TableData()
         << "<!--#status CallSignalAddresses-->"
         << PHTML::TableData("NOWRAP")
         << "<!--#status EndPointAliases-->"
         << PHTML::TableData("NOWRAP")
         << "<!--#status Application-->"
         << PHTML::TableData("align=center")
         << "<!--#status ActiveCalls-->"
       << "<!--#macroend EndPointStatus-->"
       << PHTML::TableEnd()

       << PHTML::Paragraph()

       << PHTML::TableStart("border=1")
       << PHTML::TableRow()
       << PHTML::TableHeader()
       << "&nbsp;Call&nbsp;Identifier&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;End&nbsp;Point&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Source&nbsp;Address&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Destination&nbsp;Address&nbsp;"
       << "<!--#macrostart CallStatus-->"
         << PHTML::TableRow()
         << PHTML::TableData("NOWRAP")
         << "<!--#status CallIdentifier-->"
         << PHTML::TableData("NOWRAP")
         << "<!--#status EndPointIdentifier-->"
         << PHTML::TableData("NOWRAP")
         << "<!--#status SourceAddress-->"
         << PHTML::TableData("NOWRAP")
         << "<!--#status DestinationAddress-->"
       << "<!--#macroend CallStatus-->"
       << PHTML::TableEnd()

       << PHTML::Form()
       << PHTML::HRule()

       << app.GetCopyrightText()
       << PHTML::Body();

  string = html;
}


BOOL MainStatusPage::Post(PHTTPRequest & request,
                          const PStringToString & data,
                          PHTML & msg)
{
  PTRACE(2, "Main\tClear call POST received " << data);

  PServiceHTML::ProcessMacros(request, msg, "html/status.html",
                              PServiceHTML::LoadFromFile|PServiceHTML::NoSignatureForFile);
  return TRUE;
}


static void SpliceMacro(PString & text, const PString & token, const PString & value)
{
  PRegularExpression RegEx("<?!--#status[ \t\r\n]+" + token + "[ \t\r\n]*-->?",
                           PRegularExpression::Extended|PRegularExpression::IgnoreCase);
  PINDEX pos, len;
  while (text.FindRegEx(RegEx, pos, len))
    text.Splice(value, pos, len);
}


PString MyEndPoint::OnLoadEndPointStatus(const PString & htmlBlock)
{
  PINDEX i;
  PString substitution;

  for (PSafePtr<H323RegisteredEndPoint> ep = gatekeeperServer.GetFirstEndPoint(PSafeReadOnly); ep != NULL; ep++) {
    // make a copy of the repeating html chunk
    PString insert = htmlBlock;

    SpliceMacro(insert, "EndPointIdentifier", ep->GetIdentifier());

    PStringStream addresses;
    for (i = 0; i < ep->GetSignalAddressCount(); i++) {
      if (i > 0)
        addresses << "<br>";
      addresses << ep->GetSignalAddress(i);
    }
    SpliceMacro(insert, "CallSignalAddresses", addresses);

    PStringStream aliases;
    for (i = 0; i < ep->GetAliasCount(); i++) {
      if (i > 0)
        aliases << "<br>";
      aliases << ep->GetAlias(i);
    }
    SpliceMacro(insert, "EndPointAliases", aliases);

    PString str = "<i>Name:</i> " + ep->GetApplicationInfo();
    str.Replace("\t", "<BR><i>Version:</i> ");
    str.Replace("\t", " <i>Vendor:</i> ");
    SpliceMacro(insert, "Application", str);

    SpliceMacro(insert, "ActiveCalls", ep->GetCallCount());

    // Then put it into the page, moving insertion point along after it.
    substitution += insert;
  }

  return substitution;
}


PString MyEndPoint::OnLoadCallStatus(const PString & htmlBlock)
{
  PString substitution;

  for (PSafePtr<H323GatekeeperCall> call = gatekeeperServer.GetFirstCall(PSafeReadOnly); call != NULL; call++) {
    // make a copy of the repeating html chunk
    PString insert = htmlBlock;

    SpliceMacro(insert, "CallIdentifier",     call->GetCallIdentifier().AsString());
    SpliceMacro(insert, "EndPointIdentifier", call->GetEndPoint().GetIdentifier());
    SpliceMacro(insert, "SourceAddress",      call->GetSourceAddress());
    SpliceMacro(insert, "DestinationAddress", call->GetDestinationAddress());

    // Then put it into the page, moving insertion point along after it.
    substitution += insert;
  }

  return substitution;
}


PCREATE_SERVICE_MACRO_BLOCK(EndPointStatus,P_EMPTY,P_EMPTY,block)
{
  return OpenGk::Current().GetEndPoint().OnLoadEndPointStatus(block);
}


PCREATE_SERVICE_MACRO_BLOCK(CallStatus,P_EMPTY,P_EMPTY,block)
{
  return OpenGk::Current().GetEndPoint().OnLoadCallStatus(block);
}


///////////////////////////////////////////////////////////////

MyEndPoint::MyEndPoint()
#ifdef _MSC_VER
#pragma warning(disable:4355)
#endif
  : gatekeeperServer(*this)
#ifdef _MSC_VER
#pragma warning(default:4355)
#endif
{
  terminalType = e_GatekeeperWithAudioMP;
}


MyEndPoint::~MyEndPoint()
{
  ClearAllCalls();
}


BOOL MyEndPoint::Initialise(PConfig & cfg, PConfigPage * rsrc)
{
  PINDEX i;

  // Set up software codecs.
  if (capabilities.GetSize() == 0) {
    SetCapability(0, 0, new H323_GSM0610Capability);
    SetCapability(0, 0, new H323_G711Capability(H323_G711Capability::muLaw));
    SetCapability(0, 0, new H323_G711Capability(H323_G711Capability::ALaw));

    H323_UserInputCapability::AddAllCapabilities(capabilities, 0, 1);
    PSYSTEMLOG(Info, "Main\tCodecs (in preference order):\n" << setprecision(2) << capabilities);
  }

  // Local alias name for H.323 endpoint
  SetLocalUserName(cfg.GetString(LocalUserNameKey, PProcess::Current().GetName()));
  rsrc->Add(new PHTTPStringField(LocalUserNameKey, 25, GetLocalUserName()));

  // IP TOS byte
  unsigned tos = cfg.GetInteger(TypeOfServiceKey, GetRtpIpTypeofService());
  SetRtpIpTypeofService(tos);
  rsrc->Add(new PHTTPIntegerField(TypeOfServiceKey, 0, 255, tos));

  // Interfaces to listen on
  PINDEX arraySize = cfg.GetInteger(GATEWAY_LISTENERS_SECTION, LISTENER_INTERFACE_KEY" Array Size");
  listeners.RemoveAll();
  if (arraySize == 0)
    StartListener("*:*");
  else {
    for (i = 0; i < arraySize; i++) {
      H323TransportAddress iface = cfg.GetString(GATEWAY_LISTENERS_SECTION, psprintf(LISTENER_INTERFACE_KEY" %u", i+1), "");
      if (!StartListener(iface)) {
        PSYSTEMLOG(Error, "Main\tCould not start listener: " << iface);
      }
    }
  }
  rsrc->Add(new PHTTPFieldArray(new PHTTPStringField(GATEWAY_LISTENERS_SECTION"\\"LISTENER_INTERFACE_KEY, "Gateway Interfaces", 20), FALSE));

  if (listeners.IsEmpty()) {
    PSYSTEMLOG(Fatal, "Main\tCould not open H.323 Listener");
    return FALSE;
  }

  return gatekeeperServer.Initialise(cfg, rsrc);
}


H323Connection * MyEndPoint::CreateConnection(unsigned /*callReference*/)
{
  // Need to add a creation of a connection that can do gatekeeper routing.
  return NULL;
}


///////////////////////////////////////////////////////////////

MyGatekeeperServer::MyGatekeeperServer(MyEndPoint & ep)
  : H323GatekeeperServer(ep),
    endpoint(ep)
{
}


BOOL MyGatekeeperServer::Initialise(PConfig & cfg, PConfigPage * rsrc)
{
  PINDEX i;

  PString gkid = cfg.GetString(GatekeeperIdentifierKey, OpenGk::Current().GetName() + " on " + PIPSocket::GetHostName());
  rsrc->Add(new PHTTPStringField(GatekeeperIdentifierKey, 25, gkid));
  SetGatekeeperIdentifier(gkid);

  // Interfaces to listen on
  PINDEX arraySize = cfg.GetInteger(GATEKEEPER_LISTENERS_SECTION, LISTENER_INTERFACE_KEY" Array Size");
  if (arraySize == 0)
    AddListener("*");
  else {
    for (i = 0; i < arraySize; i++) {
      H323TransportAddress iface = cfg.GetString(GATEKEEPER_LISTENERS_SECTION, psprintf(LISTENER_INTERFACE_KEY" %u", i+1), "");
      AddListener(iface);
    }
  }
  rsrc->Add(new PHTTPFieldArray(new PHTTPStringField(GATEKEEPER_LISTENERS_SECTION"\\"LISTENER_INTERFACE_KEY, "Gatekeeper Interfaces", 20), FALSE));

  availableBandwidth = cfg.GetInteger(AvailableBandwidthKey, availableBandwidth/10)*10;
  rsrc->Add(new PHTTPIntegerField(AvailableBandwidthKey, 1, INT_MAX, availableBandwidth/10, "kb/s"));

  defaultBandwidth = cfg.GetInteger(DefaultBandwidthKey, defaultBandwidth/10)*10;
  rsrc->Add(new PHTTPIntegerField(DefaultBandwidthKey, 1, INT_MAX, defaultBandwidth/10, "kb/s"));

  maximumBandwidth = cfg.GetInteger(MaximumBandwidthKey, maximumBandwidth/10)*10;
  rsrc->Add(new PHTTPIntegerField(MaximumBandwidthKey, 1, INT_MAX, maximumBandwidth/10, "kb/s"));

  defaultTimeToLive = cfg.GetInteger(DefaultTimeToLiveKey, defaultTimeToLive);
  rsrc->Add(new PHTTPIntegerField(DefaultTimeToLiveKey, 10, 86400, defaultTimeToLive, "seconds"));

  overwriteOnSameSignalAddress = cfg.GetBoolean(OverwriteOnSameSignalAddressKey, overwriteOnSameSignalAddress);
  rsrc->Add(new PHTTPBooleanField(OverwriteOnSameSignalAddressKey, overwriteOnSameSignalAddress));

  canOnlyCallRegisteredEP = cfg.GetBoolean(CanOnlyCallRegisteredEPKey, canOnlyCallRegisteredEP);
  rsrc->Add(new PHTTPBooleanField(CanOnlyCallRegisteredEPKey, canOnlyCallRegisteredEP));

  canOnlyAnswerRegisteredEP = cfg.GetBoolean(CanOnlyAnswerRegisteredEPKey, canOnlyAnswerRegisteredEP);
  rsrc->Add(new PHTTPBooleanField(CanOnlyAnswerRegisteredEPKey, canOnlyAnswerRegisteredEP));

  answerCallPreGrantedARQ = cfg.GetBoolean(AnswerCallPreGrantedARQKey, answerCallPreGrantedARQ);
  rsrc->Add(new PHTTPBooleanField(AnswerCallPreGrantedARQKey, answerCallPreGrantedARQ));

  makeCallPreGrantedARQ = cfg.GetBoolean(MakeCallPreGrantedARQKey, makeCallPreGrantedARQ);
  rsrc->Add(new PHTTPBooleanField(MakeCallPreGrantedARQKey, makeCallPreGrantedARQ));

  isGatekeeperRouted = cfg.GetBoolean(IsGatekeeperRoutedKey, isGatekeeperRouted);
  rsrc->Add(new PHTTPBooleanField(IsGatekeeperRoutedKey, isGatekeeperRouted));

  aliasCanBeHostName = cfg.GetBoolean(AliasCanBeHostNameKey, aliasCanBeHostName);
  rsrc->Add(new PHTTPBooleanField(AliasCanBeHostNameKey, aliasCanBeHostName));

  usingH235 = cfg.GetBoolean(UseH235Key, usingH235);
  rsrc->Add(new PHTTPBooleanField(UseH235Key, usingH235));

  arraySize = cfg.GetInteger(USERS_SECTION, USERS_KEY" Array Size");
  for (i = 0; i < arraySize; i++) {
    cfg.SetDefaultSection(psprintf(USERS_SECTION"\\"USERS_KEY" %u", i+1));
    PString username = cfg.GetString(UsernameKey);
    if (!username) {
      PString password = PHTTPPasswordField::Decrypt(cfg.GetString(PasswordKey));
      passwords.SetAt(username, password);
    }
  }

  PHTTPCompositeField * security = new PHTTPCompositeField(USERS_SECTION"\\"USERS_KEY" %u\\", "Authentication Credentials");
  security->Append(new PHTTPStringField(UsernameKey, 25));
  security->Append(new PHTTPPasswordField(PasswordKey, 25));
  rsrc->Add(new PHTTPFieldArray(security, FALSE));

  return TRUE;
}


H323GatekeeperCall * MyGatekeeperServer::CreateCall(const OpalGloballyUniqueID & id,
                                                    H323GatekeeperCall::Direction dir)
{
  return new MyGatekeeperCall(*this, id, dir);
}


///////////////////////////////////////////////////////////////

MyGatekeeperCall::MyGatekeeperCall(MyGatekeeperServer & gk,
                                   const OpalGloballyUniqueID & id,
                                   Direction dir)
  : H323GatekeeperCall(gk, id, dir)
{
}


H323GatekeeperRequest::Response MyGatekeeperCall::OnAdmission(H323GatekeeperARQ & info)
{
#ifdef TEST_TOKEN
  info.acf.IncludeOptionalField(H225_AdmissionConfirm::e_tokens);
  info.acf.m_tokens.SetSize(1);
  info.acf.m_tokens[0].m_tokenOID = "1.2.36.76840296.1";
  info.acf.m_tokens[0].IncludeOptionalField(H235_ClearToken::e_nonStandard);
  info.acf.m_tokens[0].m_nonStandard.m_nonStandardIdentifier = "1.2.36.76840296.1.1";
  info.acf.m_tokens[0].m_nonStandard.m_data = "SnfYt0jUuZ4lVQv8umRYaH2JltXDRW6IuYcnASVU";
#endif

#ifdef TEST_SLOW_ARQ
  if (info.IsFastResponseRequired()) {
    if (H323GatekeeperCall::OnAdmission(info) == H323GatekeeperRequest::Reject)
      return H323GatekeeperRequest::Reject;

    return H323GatekeeperRequest::InProgress(5000); // 5 seconds maximum
  }

  PTimeInterval delay = 500+PRandom::Number()%3500; // Take from 0.5 to 4 seconds
  PTRACE(3, "RAS\tTest ARQ delay " << delay);
  PThread::Sleep(delay);
  return H323GatekeeperRequest::Confirm;
#else
  return H323GatekeeperCall::OnAdmission(info);
#endif
}


// End of File ///////////////////////////////////////////////////////////////
